module Impl = struct
  open Base

  module type Params = sig
    val db: Sqlite3.db
  end

  module type S = sig
    (** Creates new entry and inserts it or updates existing one in DB *)
    val create_or_update:
      user_id:string
      -> date:Core.Date.t
      -> content:string
      -> Entry.t

    (** Finds entry written by the given user at the given date *)
    val get_by_user_id_and_date: string -> Core.Date.t -> Entry.t option

    val find_dates_with_entries: string -> int -> Core.Date.t Base.Hash_set.t
  end

  module Make(P: Params) = struct
    let get_by_user_id_and_date user_id date =
      let open Sqlite3.Data in
      let results = Oksql.exec_fetch (module Entry) P.db
          "SELECT user_id, date, content, created_at, updated_at \
           FROM diary_entries \
           WHERE user_id = :user_id \
           AND date = :date;"
          ~params:[
            ":user_id", TEXT user_id;
            ":date", TEXT (Core.Date.to_string_iso8601_basic date);
          ] in
      Array.find ~f:(Fn.const true) results

    let create_or_update ~user_id ~date ~content =
      let open Sqlite3.Data in
      let open Entry in
      let updated_at = Core.Time.now () in
      let created_at =
        get_by_user_id_and_date user_id date
        |> Option.map ~f:(fun e -> e.created_at)
        |> Option.value ~default:updated_at in
      Oksql.exec_only P.db
        "INSERT OR REPLACE INTO diary_entries \
         (user_id, date, content, created_at, updated_at) \
         VALUES \
         (:user_id, :date, :content, :created_at, :updated_at);"
        ~params:[
          ":user_id", TEXT user_id;
          ":date", TEXT (Core.Date.to_string_iso8601_basic date);
          ":content", TEXT content;
          ":created_at", TEXT (Core.Time.to_string created_at);
          ":updated_at", TEXT (Core.Time.to_string updated_at);
        ];
      { user_id; date; content; created_at; updated_at }

    module Date_from_sql = struct
      type t = Core.Date.t

      let of_sql_row = function[@warning "-4"]
        | [| "date", Sqlite3.Data.TEXT date_str |] ->
          Core.Date.of_string_iso8601_basic ~pos:0 date_str
        | row ->
          raise (Oksql.Conversion_failed row)
    end

    let find_dates_with_entries user_id year =
      let open Sqlite3.Data in
      let dates = Oksql.exec_fetch (module Date_from_sql) P.db
          "SELECT date \
           FROM diary_entries \
           WHERE user_id = :user_id \
           AND date LIKE :year_pattern;"
          ~params:[
            ":user_id", TEXT user_id;
            ":year_pattern", TEXT (Printf.sprintf "%d%%" year);
          ] in
      Core.(Hash_set.of_list (module Date) (Array.to_list dates))
  end
end

module type S = sig
  module type Params = sig
    include Impl.Params
  end

  module type S = sig
    include Impl.S
  end

  module Make(_: Params): S
end
